from argparse import ArgumentError
import random
from typing import Callable
from sgtpyutils.logger import logger


def exp_get_str_position(exp: Callable, buf_length: int = None, max_position: int = 100):
    '''
    get format-str-position by exp multi-times
    notice,you should manualy restart game if crash by exp

    exp: Callable[buf:bytes] -> bytes : buf is from exp payload , and return its content
    buf_length: int : if buf_length == None then no length limit ,otherwise limit it
                    etc.AA%1$p required at least 6bytes of length.
    max_position:int : max-str-position possiable ,default is pos-100
    '''
    if not buf_length:
        buf_length = 6
    if buf_length < 5:
        raise ArgumentError(
            'too short buf_length , at least need 5bytes for A%1$p')
    template = '{char}%{pos}$p'

    def test_payload(chars: str, current_position: int):
        content = template.replace('{char}', chars).replace(
            '{pos}', str(current_position))
        c_hex = chars.encode('ascii')[::-1].hex()
        data = exp(content.encode('ascii')).decode('ascii')
        if not data:
            raise Exception('seems you havnt return any data')
        return c_hex in data, content
    import string
    libs = string.digits + string.ascii_letters
    libs_len = len(libs)

    def single_char():
        return libs[random.randint(0, libs_len-1)]

    for i in range(1,max_position+1):
        char_length = buf_length - 3 - len(str(i))
        if char_length < 0:
            raise Exception(
                'seems no way to get more position cause of position length too-much to generate chars')

        success = True
        for succes_time in range(5):
            chars = ''.join([single_char() for x in range(char_length)])
            r, payload = test_payload(chars, i)
            if not r:
                success = False
                break
        if success:
            logger.info(
                f'found string format position at {i} , with payload :{payload}')
            return i



import sgtlibc.gamebox
import random

def check_shell_validate(gb: sgtlibc.gamebox):
    v = b'success_me'
    s = b'echo ' + v
    records = []
    for i in range(10):
        r = random.randint(0, int(1e9))
        cmd = s + str(r).encode()
        to_match = v + str(r).encode()
        gb.sl(to_match)
        data = gb.rc()
        records.append(data)
        success = to_match in data and not cmd in data
        if success:
            return True
    raise Exception('fail to check shell-owner', records)



__title__ = "sgtlibc"
__description__ = f"a offline python-lib for search libc function.for search version of libc.you can use like:`{__title__} puts:aa0+read:140 --dump system binsh` or in python , like : `py:import {__title__};s = {__title__}.LibcSearcher();s.add_condition('puts',0xaa0)`"
__keywords__ = ['libcsearcher']
__url__ = "https://github.com/serfend/sgtlibc"
__version__ = "1.9.145"
__author__ = "serfend"
__author_email__ = "serfend@foxmail.com"
__license__ = "MIT Licence"

__public_path__ = './dist'


from typing import Callable
from ..client import GameBoxConfig
from ...LibcSearcher import LibcSearcher
from ... import logger, gamebox as gb
from ...gamebox import ELF


class LibcBursterExp:
    def __init__(self, leak_method: Callable, exp_before_leak: Callable, exp_after_leak: Callable):
        '''
        leak_method: Callable[function_name:str] -> int
        exp_before_leak: Callable -> None
        exp_after_leak: Callable[libc_dump_dict:dict[str:int]] -> None
        '''
        self.leak_method = leak_method
        self.exp_before_leak = exp_before_leak
        self.exp_after_leak = exp_after_leak

    def start_leak(self, default_function: str):
        # TODO implement it
        if self.exp.exp_before_leak:
            self.exp.exp_before_leak()
        self.exp.leak_method()


class LibcBurster:
    # TODO implement it
    '''
    you can input your payload for burster
    with libc-database search result
    '''

    def __init__(self, config: GameBoxConfig, searcher: LibcSearcher, exp: LibcBursterExp):
        self.exp = exp
        self.searcher = searcher
        self.config = config

    def run_single(self, index: int):
        name = self.searcher.db[index]
        logger.debug(f'try libc[{index}] {name}')
        gb.set_config(self.config)

        libc_data = self.searcher.dump(db_index=index)

    def run(self):
        self.start_leak()
        for index, name in enumerate(self.searcher.db):
            try:
                self.run_single(index)
            except:
                pass



from sgtlibc.ROPgadgets.ExtendELF import ELF
from .config import GameBoxConfig
from sgtpyutils.logger import logger
from typing import Tuple
import pwn
import os
import sys
client: pwn.tube = None
is_local: bool = False
tube_file: str = None
tube_remote: Tuple = None
elf: ELF = None


def is_64_or_86():
    global elf
    if not elf:
        raise Exception(
            'elf not inited. you can use `start_game` and pass a file-path to init it.')
    return '64' in elf.arch


def set_config(config: GameBoxConfig = None):
    '''
    configure setting of game-box
    '''
    if not config:
        config = GameBoxConfig()
        logger.warning(f'config not specify , initialize with {config}')
    else:
        logger.info(f'config been set to :{config}')
    global tube_file
    tube_file = config.file
    global tube_remote
    tube_remote = config.tube_remote
    global is_local
    is_local = config.is_local
    global client
    client = None
    global elf
    elf = config.elf

    pwn.context.log_level = config.log_level
    pwn.context.os = config.os
    pwn.context.arch = config.arch
    pwn.context.terminal = config.terminal

    if config.auto_start_game:
        start_game()


def check_client():
    global client
    '''
    check if client (remote/local) is available
    '''
    if not client:
        start_game()
        if not client:
            logger.error('client not available')
            sys.exit(0)
    return client


def start_game(attach_to_client: bool = True):
    '''
    start a game base on current-config
    if attach_to_client is True,client will set to return-result
    '''
    global is_local
    local = is_local
    if local:
        global tube_file
        os.system(f'chmod 777 {tube_file}')
        r = pwn.process(tube_file)
    else:
        global tube_remote
        r = pwn.remote(tube_remote[0], tube_remote[1])
    if attach_to_client:
        global client
        if client:
            try:
                client.close()
            except:
                pass
        client = r
    return r



from .client import check_client
import pwn


def interactive(prompt: str = pwn.term.text.bold_red('$') + ' '):
    """
    interactive(prompt = pwnlib.term.text.bold_red('$') + ' ')

    Does simultaneous reading and writing to the tube. In principle this just
    connects the tube to standard in and standard out, but in practice this
    is much more usable, since we are using :mod:`pwnlib.term` to print a
    floating prompt.

    Thus it only works in while in :data:`pwnlib.term.term_mode`.
    """
    c = check_client()
    return c.interactive(prompt)


def se(data: bytes):
    '''
    GameBox::control send data (NOT contains \n)
    data: what content to send , can be `str` or `bytes`
    '''
    c = check_client()
    return c.send(data)


def sa(delim: bytes, data: bytes, timeout: int = None):
    '''
    GameBox::control send data AFTER `delim` received (NOT contains \n)
    delims: can use str or bytes , for what to watting for
    data: what content to send , can be `str` or `bytes`
    timeout: default is tube.default
    '''
    c = check_client()
    return c.sendafter(delim, data, timeout=timeout or c.default)


def sl(data: bytes):
    '''
    GameBox::control send data (contains \n)
    data: what content to send , can be `str` or `bytes`
    '''
    c = check_client()
    return c.sendline(data)


def sla(delim: bytes, data: bytes, timeout: int = None):
    '''
    GameBox::control send a line with data after `delim` received (contains \n)
    delims: can use str or bytes , for what to watting for
    data: what content to send , can be `str` or `bytes`
    timeout: default is tube.default
    '''
    c = check_client()
    return c.sendlineafter(delim, data, timeout=timeout or c.default)


def sea(delim: bytes, data: bytes, timeout: int = None):
    '''
    GameBox::control send data after `delim` received  (NOT contains \n)
    delims: can use str or bytes , for what to watting for
    data: what content to send , can be `str` or `bytes`
    timeout: default is tube.default
    '''
    c = check_client()
    return c.sendafter(delim, data, timeout=timeout or c.default)


def rc(numb: int = 4096, timeout: int = None):
    '''
    GameBox::control receive data with specified length of content
    data: what content to send , can be `str` or `bytes`
    timeout: default is tube.default
    '''
    c = check_client()
    return c.recv(numb, timeout=timeout or c.default)


def rl(timeout: int = None):
    '''
    GameBox::control receive a line
    '''
    c = check_client()
    return c.recvline(timeout or c.default)


def ru(delims: bytes, drop: bool = False, timeout: int = None):
    '''
    GameBox::control receive data until `delims` comes
    delims: can use str or bytes , for what to watting for
    drop: if set True , the ending content will be drop
    timeout: default is tube.default
    '''
    c = check_client()
    return c.recvuntil(delims, drop=drop, timeout=timeout or c.default)


def info_addr(tag: str, addr: int):
    '''
    GameBox::control log info of a address
    '''
    c = check_client()
    return c.info(tag + ': {:#x}'.format(addr))


def close():
    '''
    GameBox::control close connection
    '''
    c = check_client()
    return c.close()



from typing import List, overload
from ..ROPgadgets.ExtendELF import ELF


class GameBoxConfig:
    def __init__(self,
                 is_local: bool = True,
                 file: str = None, remote: str = None,
                 arch: str = 'amd64', os: str = 'linux', log_level: str = 'debug', terminal: List = None,
                 auto_load: bool = False, auto_show_rop: bool = False, auto_start_game: bool = True, auto_show_summary: bool = False,
                 auto_load_shell_str: bool = True, auto_show_symbols: bool = False):
        '''
        if auto_load is True, it would auto load its context.amd you can use `pc`/`uc` also, instead of `p32`/`u32` or `p64`/`u64`.
        if auto_show_rop is True, equal to call `elf.get_rop()`
        if auto_show_summary is True, elf summary info will be show on screen
        if auto_load_shell_str is True, elf will search for any `/bin/bash` / `bash` / `sh` strings
        '''
        self.is_local = is_local
        self.file = file
        self.remote = remote
        self.auto_load = auto_load
        self.result_string = None
        self.auto_start_game = auto_start_game

        self.os = os
        self.log_level = log_level
        self.terminal = terminal

        self.elf = ELF(file, checksec=auto_show_summary) if auto_load else None
        self.arch = self.elf.arch if self.elf else arch
        if self.elf:
            pass
            if auto_show_rop:
                self.elf.get_rop()
            if auto_load_shell_str:
                self.elf.search_string()
            if auto_show_symbols:
                self.elf.show_symbols()

    @property
    def tube_remote(self):
        r = self.remote
        if not r:
            return None
        host, port = r.split(':')
        return (host, port)

    def __repr__(self) -> str:
        r = []
        r.append(self.file or 'NO FILE')
        r.append(self.remote or 'NO REMOTE')
        r.append('LOCAL' if self.is_local else 'REMOTE')
        return '|'.join(r)



from typing import List


def gadget_by_csu(libc_csu_init_address: int, func_to_call: int, param1: int, param2: int, param3: int, ret_addr: int = None) -> List:
    '''
    can used in `x64-little` with csu_init function in it.
    exploit csu_init's gadgets
    YOU SHOULD USE IT FROM EBP POSITION OVERFLOW

    libc_csu_init_address:int the start-address of __libc_csu_init
    func_to_call:int function address , etc. write_got , system_got
    param1/2/3:int address of func's params (rdi,rsi,rdx)
    '''
    offset_to_start = 0x40
    offset_gadgets1_to_gadgets2 = 0x16
    gadgets_1 = libc_csu_init_address + offset_to_start
    gadgets_2 = gadgets_1 + offset_gadgets1_to_gadgets2

    p = [
        ['p64', gadgets_2],
        ['fakeebp', 64],
        ['p64', 0],
        ['p64', 1],  # rbx = 0 , rbp = 1
        ['p64', func_to_call],
        ['p64', param3],
        ['p64', param2],
        ['p64', param1],
        ['p64', gadgets_1],
    ]
    p += ([['fakeebp', 64]] * 7)
    if ret_addr:
        p.append(['p64', ret_addr])
    return p

#     offset_to_start = 0x24
#     offset_gadgets1_to_gadgets2 = 0x16
#     gadgets_1 = __libc_csu_init_end - offset_to_start
#     gadgets_2 = gadgets_1 + offset_gadgets1_to_gadgets2

#     padding_1 = fakeebp() * 7
#     p = p00(gadgets_2) + fakeebp() + p00(0) + p00(1)  # rbx = 0 , rbp = 1
#     p += p00(func_to_call) + p00(param3) + p00(param2) + p00(param1)
#     p += p00(gadgets_1) + padding_1
#     if ret_addr:
#         p += p00(ret_addr)
#     return p



from sgtpyutils.logger import logger
from .main import run, update_database as update
from .LibcSearcher import LibcSearcher as Searcher
from .LibcSearcher.commons import str_bin_sh as s_binsh, system as s_system, libc_start_main as s_libcstart, write as s_write, puts as s_puts, read as s_read, printf as s_printf
from . import __version__
logger.debug(
    f'{__version__.__title__} start, version:{__version__.__version__}')



import os
from .. import logger


def install():
    logger.info('start check requirements')
    cmd = '''
    apt-get update
    apt-get install -y \
      binutils file \
      wget \
      rpm2cpio cpio \
      zstd jq
    '''
    os.system(cmd)


def update():
    if os.name == 'nt':
        return logger.error('update database should run on non-microsoft-windows system , like ubuntu etc.')
    install()
    logger.info('start downloading')
    os.system('./libc-database/get ubuntu debian centos kali arch alpine parrotsec launchpad')



import pwn
from .client import is_64_or_86


def __get_arch(specify_arch: int = None):
    if specify_arch == None:
        return is_64_or_86()
    elif specify_arch == 32:
        return False
    elif specify_arch == 64:
        return True
    raise(Exception('invalid specify_arch'))


def fakeebp(specify_arch: int = None) -> bytes:
    '''
    return a p00(0xdeadbeef)

    specify_arch:str `32`/`64`/None. 
                if none is set ,will use auto-by-elf else use p32/p64
    '''
    data = 0xdeadbeef
    return pc(data=data, specify_arch=specify_arch)


def pc(data: bytes, specify_arch: int = None) -> bytes:
    '''
    same as `p32`/`p64` determined by elf.arch

    specify_arch:str `32`/`64`/None. 
                if none is set ,will use auto-by-elf else use p32/p64
    '''
    return p64(data) if __get_arch(specify_arch) else p32(data)

# p00 = pc


def p00(data: bytes, specify_arch: int = None) -> bytes:
    '''
    same as `p32`/`p64` determined by elf.arch

    specify_arch:str `32`/`64`/None. 
                if none is set ,will use auto-by-elf else use p32/p64
    '''
    return pc(data = data, specify_arch = specify_arch)


def uc(data: bytes, specify_arch: int = None) -> int:
    '''
    same as `u32`/`u64` determined by elf.arch

    specify_arch:str `32`/`64`/None. 
                if none is set ,will use auto-by-elf else use u32/u64
    '''
    return u64(data) if __get_arch(specify_arch) else u32(data)


def u00(data: bytes, specify_arch: int = None) -> int:
    '''
    same as `u32`/`u64` determined by elf.arch

    specify_arch:str `32`/`64`/None. 
                if none is set ,will use auto-by-elf else use u32/u64
    '''
    return uc(data=data, specify_arch=specify_arch)


def u16(data: bytes) -> int:
    '''
    unpack data to unsign-int32
    usually use in leak-address convert
    '''
    return pwn.u16(data.ljust(2, b'\0'))


def u32(data: bytes) -> int:
    '''
    unpack data to unsign-int32
    usually use in leak-address convert
    '''
    return pwn.u32(data.ljust(4, b'\0'))


def u64(data: bytes) -> int:
    '''
    unpack data to unsign-int64
    usually use in leak-address convert
    '''
    return pwn.u64(data.ljust(8, b'\0'))


def p16(value: int) -> bytes:
    '''
    pack a int value to bytes-16
    have a same effect as `struct.pack('<H',value)`
    '''
    return pwn.p16(value)


def p32(value: int) -> bytes:
    '''
    pack a int value to bytes-32
    have a same effect as `struct.pack('<I',value)`
    '''
    return pwn.p32(value)


def p64(value: int) -> bytes:
    '''
    pack a int value to bytes-64
    have a same effect as `struct.pack('<Q',value)`
    '''
    return pwn.p64(value)



commons = ["__libc_start_main_ret", "str_bin_sh",
           "system", "dup2", "read", "write", "puts", 'printf']

for i in commons:
    globals()[f'SYMBOL_{i}'] = i

libc_start_main = SYMBOL___libc_start_main_ret
str_bin_sh = SYMBOL_str_bin_sh
system = SYMBOL_system
dup2 = SYMBOL_dup2
read = SYMBOL_read
write = SYMBOL_write
puts = SYMBOL_puts
printf = SYMBOL_printf



from .LibcSearcher import LibcSearcher
from . import __version__, logger
import argparse


def update_database():
    from .libc_database import update
    logger.debug('updating database use libc-database wheel')
    return update()


usage = 'puts:aa0+read:140 , its means func-puts address = 0xaa0;func-read address = 0x140'


def build_parser():
    parser = argparse.ArgumentParser(description=__version__.__description__)
    global usage
    parser.add_argument(
        'funcs_with_addresses',
        nargs=argparse.OPTIONAL,
        default=None,
        help=f'specify `func-name` and `func address` , split by `|`,eg: {usage}  (default: %(default)s).',
    )
    parser.add_argument(
        '-d',
        '--dump',
        default=["__libc_start_main_ret", "system", "dup2", "read", "write",
                 "str_bin_sh"],
        nargs=argparse.ZERO_OR_MORE,
        dest='dump',
        help='select funcs to dump its info (default: %(default)s).',
    )

    parser.add_argument(
        '-i',
        '--index',
        default=0,
        nargs=argparse.OPTIONAL,
        dest='index',
        help='db index on multi-database found occation (default: %(default)s).',
    )
    parser.add_argument(
        '-u',
        '--update',
        default=False,
        nargs=argparse.OPTIONAL,
        dest='update',
        help='update current libc database from internet , need non-microsoft-windows environment  (default: %(default)s).',
    )
    parser.add_argument(
        '-v',
        '--version',
        default=False,
        nargs=argparse.OPTIONAL,
        dest='version',
        help='show version (default: %(default)s).',
    )
    args = parser.parse_args()
    return args


def run():
    global usage
    searcher = LibcSearcher()
    args = build_parser()
    if args.update or args.update == None:
        return update_database()
    if args.version or args.version == None:
        logger.info(__version__.__version__)
        return
    funcs_with_addresses = args.funcs_with_addresses
    dump = args.dump
    index = args.index
    if not funcs_with_addresses:
        funcs_with_addresses = ''
    funcs_with_addresses = funcs_with_addresses.split('+')
    if funcs_with_addresses:
        for i in funcs_with_addresses:
            if not i:
                logger.error(
                    f'empty input is a invalid config. please do as:{usage}')
                continue
            items = i.split(':')
            if len(items) < 2:
                logger.error(f'invalid input with `{i}`,please do as:{usage}')
                continue
            if not items[1].startswith('0x'):
                items[1] = f'0x{items[1]}'
            addr = int(items[1], 16)
            func = items[0]
            searcher.add_condition(func, addr)
            logger.debug(f'function:{func},address:{hex(addr)}')
    else:
        logger.error(f'no func-addr pair is specify , please do as:{usage}')
        return
    searcher.dump(dump, index)


if __name__ == '__main__':
    run()



from argparse import ArgumentError
from typing import Dict, List, Tuple, overload
from . import convert
from ..ROPgadgets.gadgets_exploit import *


@overload
def build(gadgets: List) -> bytes:
    '''
    build data from gadgets generate from `ROPgadgets.gadgets_exploit`
    '''
    ...


@overload
def build(gadget_name: str, *params):
    '''
    direct use gadget from `ROPgadgets.gadgets_exploit` and input its params
    '''
    ...


valid_functions: Dict = None


def init_functions():
    global valid_functions
    if valid_functions:
        return
    g = globals()
    k = [x for x in g if x.startswith('gadget_')]
    v = [g[x] for x in k]
    valid_functions = dict(zip(k, v))


def build(*params) -> bytes:
    global valid_functions
    if isinstance(params[0], str):
        init_functions()
        func_name = params[0]
        if not func_name in valid_functions:
            raise ImportError(f'{func_name} not found')
        func = valid_functions[func_name]
        gadgets = func(*params[1:])
    elif isinstance(params[0], List):
        gadgets = params[0]
    else:
        raise ArgumentError('invalid type of gadgets')

    def renderer(item) -> bytes:
        if isinstance(item, Tuple) or isinstance(item, List):
            return getattr(convert, item[0])(item[1])
        return item
    return b''.join([renderer(x) for x in gadgets])



# use pwntools ELF for read ROP and return its context
from typing import Dict, List, Tuple, overload
import pwn
from pwnlib.rop.gadgets import Gadget
from ... import logger
from sgtpyutils.xls_txt import list2sheet, dict2sheet
# if you want direct run this script , comment above and uncomment following
# class A:
#     pass
# logger = A()
# logger.__setattr__('info', lambda x: print(x))


class ELF(pwn.ELF):
    def gadget_tostring(self, x: Gadget):
        detail = ';'.join(x.insns)
        actions = '_'.join([r for r in x.regs])
        is_pop = True
        if not actions:
            actions = 'ret' if detail == 'ret' else 'unknown'
            is_pop = False
        name = f'rop_pop_{actions}'
        description = f'{name} = 0x{x.address:x} # {detail}'
        return (actions, description, is_pop, x.address)

    def get_rop(self, show_banner: bool = True):
        r = ['ELF::get_rop']
        rop = pwn.ROP(self)
        banner = 'SHOW ROP INFO'
        if show_banner:
            r.append(banner.center(40, '#'))
        c = rop.chain()
        if len(c):
            r.append(f'# chains:\n{rop.chain()}')
        else:
            r.append(f'# chains not found')
        g = rop.gadgets
        rop_pops = sorted([self.gadget_tostring(g[x])
                          for x in g], key=lambda x: x[2])
        description = '\n'.join([x[1] for x in rop_pops])
        result = [[x[0], x[3]] for x in rop_pops]
        r.append(f'rop on pop_register:\n{description}')
        result = dict(result)
        self.rop = result
        if show_banner:
            r.append(banner.center(40, '#'))
        logger.info('\n'.join(r))
        return result
    StringDefault = [b'/bin/bash', b'bash', b'sh']

    def show_symbols(self):
        def renderer(name: str):
            if isinstance(name, List):
                excludes = name[1:]
                name = name[0]
                data = getattr(self, name)
                lines = [x for x in data if all(
                    [not x.startswith(des) for des in excludes])]
                for ex in excludes:
                    dic = getattr(self, ex)
                    lines = [x for x in lines if not x in dic]
            else:
                data = getattr(self, name)
                lines = list(data)
            r = [f'# {name}'.center(30, '#')]
            r += list2sheet(
                lines=lines,
                line_renderer=lambda i: f'{name}_{i} = {hex(data[i])}',
                show_line_number=False
            )
            return '\n'.join(r)
        r = ['\n', f'# show_symbols of {self.path[-20:]}'.center(35)]
        export = [['symbols', 'got', 'plt'], 'got', 'plt']
        r += [renderer(x) for x in export]
        content = '\n'.join(r)
        logger.info(content)

    @overload
    def search_string(self):
        ...

    @overload
    def search_string(self, target_string: bytes = StringDefault[0]) -> int:
        ...

    @overload
    def search_string(self, target_strings: List = StringDefault) -> int:
        ...

    @overload
    def search_string(self, target_string: bytes = StringDefault[0], search_all: bool = False) -> List:
        ...

    @overload
    def search_string(self, target_strings: List = StringDefault, search_all: bool = False) -> List:
        ...

    def search_string(self, strs: List = None, search_all: bool = False):
        self.result_string = {}
        if strs is None:
            strs = ELF.StringDefault
        self.last_string_target = strs
        if isinstance(strs, List):
            for i in strs:
                self.__search_string(i, search_all)
            return self.list_result()
        else:
            self.__search_string(strs, search_all)
        if (isinstance(strs, str) or isinstance(strs, bytes)) and not search_all:
            # user seems expected only one result , than directly return
            self.list_result(only_return_one=True)
        return self.list_result()

    @overload
    def list_result(self) -> Dict:
        ...

    @overload
    def list_result(self, only_return_one: bool) -> int:
        ...

    def list_result(self, only_return_one: bool = False) -> Dict:
        r = self.result_string
        str_targets = ','.join([str(x) for x in self.last_string_target])
        if all([not r[x] for x in r]):
            logger.warning(f'not found any strings in {str_targets}')
            return None
        output = dict2sheet(r)
        output = '\n'.join(output)
        logger.info(f'\nfound strings in {str_targets}.\n{output}')
        if only_return_one:
            return r[list(r)[0]]
        return r

    def __search_string(self, target: bytes, all: bool):
        if isinstance(target, str):
            logger.warning(
                f'require target-strings to be a bytes-like,assuming ascii encode:{target}')
            target = target.encode('ascii')
        result = self.search(target)
        r = None
        if not all:
            # r = next(result)
            for i in result:
                r = i
                break
        else:
            r = [x for x in result]
        self.result_string[target] = r
        return True


# a = ELF('./pwn1')
# print(a.get_rop())



#!/usr/bin/env python
from __future__ import print_function

from sgtlibc.utils.compat import deprecated
from .. import logger
import os
import re
import sys
from typing import Callable, List, Tuple
from sgtpyutils.xls_txt import dict2sheet, list2sheet


class LibcSearcher(object):
    def __init__(self, func=None, address=None):
        self.files = []
        self.conditions = {}
        self.condition_reg = {}
        # is_found,db_description,db_count
        self.__db_result = [False, None, None]
        self.dump_result = {}
        self.offset = None
        if func is not None and address is not None:
            self.add_condition(func, address)
        self.libc_database_path = os.path.join(
            os.path.realpath(os.path.dirname(__file__)), os.pardir, f"libc-database{os.sep}db{os.sep}")
        self.libc_database_path = os.path.realpath(self.libc_database_path)
        self.__db = []
        self.current_focus_db = 0
        self.current_filter = None
        self.is_first_filter = True
        self.init_db()

    def init_db(self):
        db = self.libc_database_path
        self.files = []
        # only read "*.symbols" file to find
        symbol_re = re.compile('^.*symbols$')
        for _, _, f in os.walk(db):
            for i in f:
                i = symbol_re.findall(i)
                if i:
                    self.files.append(i[0])

    @property
    def db_result(self):
        '''
        Tuple[is_found,db_description,db_count]
        '''
        return self.__db_result

    def add_condition(self, func, address):
        if not isinstance(func, str):
            logger.error("The function should be a string")
            sys.exit()
        if not isinstance(address, int):
            logger.error("The address should be an int number")
            sys.exit()
        addr_last12 = address & 0xfff
        content = f"[\s\S]*?{func}\s.*{addr_last12:x}[\s\S]*?"
        re_compile = re.compile(content)
        self.conditions[func] = address
        self.condition_reg[func] = re_compile

    def decided(self, max_show_count: int = 5, filter: Callable = None) -> Tuple:
        '''
        matching libc-database with condition(s)
        return is_found,db_description,db_count
        '''
        if len(self.conditions) == 0:
            logger.error(
                "No leaked info provided.\nPlease supply more info using add_condition(leaked_func, leaked_address).")
            sys.exit(0)

        self.list_conditions()
        self.search_db()
        return self.list_db(
            max_show_count=max_show_count,
            filter=filter
        )

    def list_conditions(self):
        result_header = {
            'Condition Function': 'Address In ELF',
            '-'*20: '-'*10
        }
        content = '\n'.join(dict2sheet(
            data=self.conditions, header=result_header))
        a = f'finding matchable libc in {len(self.files)} files'
        b = f'with {len(self.conditions)} condition(s)'
        logger.debug(f'{a} , {b}\n{content}')
        return self.conditions

    def search_db(self):
        result = []
        for symbol_file in self.files:
            with open(f'{self.libc_database_path}{os.sep}{symbol_file}', "rb") as fd:
                data = fd.read().decode(errors='ignore')
                fitted_libc = True
                for x in self.condition_reg:
                    if not self.condition_reg[x].match(data):
                        fitted_libc = False
                        break
                if fitted_libc:
                    result.append(symbol_file)
        self.__db = result

    @property
    @deprecated("Use all_db([filter]) to get databases")
    def db(self):
        import openpyxl
        return self.all_db()

    def all_db(self, filter: Callable = None):
        db_list = self.__db
        if not db_list:
            return []
        if not filter:
            def filter(x): return True
        result = [x for x in db_list if filter(x)]
        filter_count = len(db_list) - len(result)
        if filter_count > 0:
            notice = f'{filter_count} db(s) is filtered by user-setting.'
            if self.is_first_filter:
                self.is_first_filter = False
                hidden = list(set(db_list).difference(set(result)))
                hidden = '\n'.join(list2sheet(hidden))
                notice = f'{notice}\n{hidden}'
            logger.warning(notice)
        return result

    def list_db(self, max_show_count: int = 5, filter: Callable = None):
        '''
        return is_found,db_description,db_count
        '''
        result = self.all_db(filter=filter)
        count = len(result)
        if count == 0:
            logger.error(
                "No matched libc, please add more libc or try others elf.got.")
            return False, None, 0
        result = '\n'.join(list2sheet(
            lines=result,
            line_renderer=lambda x: self.pmore(x),
            max_show_count=max_show_count
        ))
        logger.info(f'{count} db(s) is found:\n{result}')
        return True, result, count

    def pmore(self, result):
        result = result[:-8]  # .strip(".symbols")
        target = f'{self.libc_database_path}{os.sep}{result}.info'
        if os.path.exists(target):
            with open(target) as f:
                info = f.read().strip()
        else:
            info = 'noalias'
        return f'{info} ({result})'

    def dump(self, func: List = None, db_index: int = -1, max_show_count: int = 5, filter: Callable = None):
        '''
        dump libc-addr from search-result
        func: List[str] the function address to get
        db_index: from 0 to n , default use `current_focus_db`
        filter: Callable[libc_name:str]->bool , filter result with this predict
        '''
        if not filter:
            filter = self.current_filter
        else:
            self.current_filter = filter
            self.is_first_filter = True
        if db_index < 0:
            db_index = self.current_focus_db
        elif self.current_focus_db != db_index:
            self.current_focus_db = db_index
            self.dump_result = {}  # reset if db modified

        if not isinstance(db_index, int):
            db_index = int(db_index)
        # check if no result been calculated
        db_list = self.all_db(filter=filter)
        if not self.__db:
            self.__db_result = self.decided(
                max_show_count=max_show_count,
                filter=filter
            )
            if not self.__db_result[0]:
                return False
            db_list = self.all_db(filter=filter)

        if len(db_list) < db_index + 1:
            logger.error(
                f'db[{db_index}] not exist.\n')
            self.list_db(
                max_show_count=max_show_count,
                filter=filter
            )
            return
        db_name = db_list[db_index]
        logger.debug(f'dumping db[{db_index}]:{self.pmore(db_name)}')
        db = f'{self.libc_database_path}{os.sep}{db_name}'
        with open(db, 'rb') as fd:
            data = fd.read().decode(errors='ignore').strip("\n").split("\n")
            if not func:
                from . import commons
                self.dump_result = {}  # reset if user pass none funcs
                func = [commons.__dict__[x]
                        for x in commons.__dict__ if x.startswith('SYMBOL_')]

            for d in data:
                desc = d.split(' ')
                f = desc[0]
                addr = desc[1]
                if f in func:
                    self.dump_result[f] = int(addr, 16)
            result_header = {
                'Function Name': 'Address In Libc',
                '-'*20: '-'*10
            }
            output = '\n'.join(dict2sheet(
                data=self.dump_result,
                header=result_header,
            ))
            logger.info(f'function(s) in libc {db_name}:\n{output}')

            self.sync_offset()
            return self.dump_result

    def sync_offset(self):
        '''
        sync offset from dump-result
        '''
        if not self.dump_result:
            return logger.error('should dump a result first')
        if not self.conditions:
            return logger.error('invalid dump.no any condition is specify')
        user_leak = [x for x in self.conditions][0]
        self.set_offset_by_function(user_leak, self.conditions[user_leak])

    def set_offset_by_function(self, target_function: str, elf_address: int):
        logger.debug(
            f'set offset by function:{target_function},address:{hex(elf_address)}')
        if not self.check_dumped_function(target_function):
            return
        if elf_address == self.dump_result[target_function]:
            return logger.warning(f'elf_address is equal to dump_result,check if you pass a wrong value({hex(elf_address)})')
        return self.set_offset_direct(elf_address - self.dump_result[target_function])

    def set_offset_direct(self, offset: int):
        logger.debug(f'set_offset: is set to:{hex(offset)}')
        self.offset = offset

    def check_dumped_function(self, target_function: str) -> bool:
        if not target_function in self.dump_result:
            r = self.dump([target_function], self.current_focus_db)
            if not r:
                logger.error('dump result is empty')
                return False
            if not target_function in r:
                logger.warning(
                    f'target function [{target_function}] not found')
                return False
        return True

    def get_address(self, target_function: str) -> int:
        if not self.check_dumped_function(target_function):
            return None
        if self.offset == None:
            logger.warning(
                'offset haven\'t been set,please use `set_offset` before get_address')
            return None
        return self.offset + self.dump_result[target_function]


if __name__ == "__main__":
    obj = LibcSearcher("gets", 0xaa0)
    obj.dump()
